%Monte-Carlo based localization
%----
%Algoritmo Monte Carlo per la localizzazione di robot mobili.
%Il presente lavoro  compatibile sia con tutti i prodotti compatibili con
%la dll di interfaccia ARIA.
%INPUT:
%- map : matrice di linee che descrive la mappa, 
%           map : (nLines x [x1,x2,y1,y2])
%- Corners : Matrice contenente le coordinate dei punti di angolo della
%           mappa, Corners : (nCorners x [x,y])
%- Areas : Matrice contenente la descrizione delle aree entro cui il robot
%           pu localizzarsi, Areas : (nAreas x [x1,y1,x2,y2])
%- NParticles : Numero di particelle massime da impiegare nella
%           localizzazione del robot, NParticles : (1x1, integer)
%- NSteps : Numero di passi massimi per la simulazione, nel caso il robot
%           si localizzi prima di NSteps verr cmq inseguita la posizione 
%           del robot, NSteps : (1x1, integer)
%- TypeOfResample : Indica la tipologia di campionamento da impiegare:
%           0 : Campionamento con sovrascrittura
%           1 : Campionamento con sovrascrittura e ridimensionamento numero
%           di particelle
%           Si rimanda alla documentazione cartacea per i dettagli di ogni
%           tipo di campionamento.
%- TypeOfInput : Indica la sorgente dei comandi da impartire al robot
%           0 : obstacle avoidance autonomo
%           1 : input manuale dei comandi
%- AviFileName : Nome del file Avi su cui registrare il filmato della localizzazione,
%           AviFileName : (1x1, string)
%           Se il campo viene lasciato vuoto non viene creato alcun filmato dell'elaborazione
%OUTPUT: 
%   void
%----
%esempio:
%MCL_v3('-rp COM1',Dipartimento,DipCorner,DipAreasMinus,100,50,1,0,'prova.avi')
%----
%Powered by Mancini Adriano & Caponetti Fabio A.A. 2004/2005
%CdL Specialistica in Automazione Industriale
function [xEstTot, xOdom, nPart] = MCL(Robot_address,map, Corners, Areas, NParticles, nSteps, TypeOfResample, TypeOfInput, AviFileName)
global Map; global corners; global nParticles;

Map = map;
corners = Corners;
nParticles = NParticles;

close all;


try
    r=robot(Robot_address);
catch
    disp 'Connection Failed!'
    return
end

%FIXME: rimuovere le note dal try catch v ripristinato sino al segno
try
    setabsolutemaxrotvel(r, 40);
    if nargin < 9
        [xEstTot,xOdom,nPart] = localization(r,Areas,nSteps,TypeOfResample,TypeOfInput,'');
    else
        [xEstTot,xOdom,nPart] = localization(r,Areas,nSteps,TypeOfResample,TypeOfInput,AviFileName);
    end
catch
    lasterr
    setvel(r, [0, 0]);
    disconnect(r);
    return;
end
%fine ripristino
stop(r);
disconnect(r);


%---%
%Monte Carlo Localization routine
function [xEstTot, xOdom, nPart] = localization(r,Areas,nSteps,TypeOfResample,TypeOfInput,AviFileName)
global Map; global corners; global nParticles; global xP;

xEstTot = [];
xOdom = [];
nPart= [];

xOdomLast = getpose(r);
xOdomLast(3) = angle_wrap(deg2rad(xOdomLast(3)));

xMap=max(max(Map(:,1:2)));
yMap=max(max(Map(:,3:4)));

xP = []; %zeros(nParticles,3);
%La distribuzione iniziale delle particelle  creata secondo una
%distribuzione di probabilit uniforme
%N.B. L'orientamento del robot  compreso tra [0;2pi]

nAreas = size(Areas,1);

xMax = max(Areas(:,1),Areas(:,3));
yMax = max(Areas(:,2),Areas(:,4));

xMin = min(Areas(:,1),Areas(:,3));
yMin = min(Areas(:,2),Areas(:,4));

Areas = [ xMin xMax yMin yMax ];

DistribuiteParticlesOrientation(Areas, 3);

xP(:,4) = 1/nParticles; %ogni particella ha lo stesso peso d'apprincipio

%Inizializzo la grafica
figure(1); hold on; grid off; axis equal;
DrawMap(Map);hold on;
set(gcf,'doublebuffer','on');
hPoints = plot(xP(:,1),xP(:,2),'.');
meanXp = plot(0,0,'r*');
drawnow;

noAvi = 0;
if AviFileName == ' '
    noAvi = 1;
else
    try
        mov = avifile(AviFileName);
        F = getframe(gca);
        mov = addframe(mov,F);
    catch
        disp ('Problemi nella creazione del file avi...');
        noAvi = 1;
    end
end
translation = 300;
for k = 2:nSteps
    k
    switch TypeOfInput
        case 0
            RealSonar = readsonar(r);
            RealSonar = RefineSonarMesure(RealSonar);
            th = 0;
            translation = 300;
            SecurityThreshold = 400;

            if      (RealSonar(1) <= SecurityThreshold) || (RealSonar(2) <= SecurityThreshold)||...
                    (RealSonar(3) <= SecurityThreshold) || (RealSonar(4) <= SecurityThreshold)||...
                    (RealSonar(5) <= SecurityThreshold) || (RealSonar(6) <= SecurityThreshold)||...
                    (RealSonar(7) <= SecurityThreshold) || (RealSonar(8) <= SecurityThreshold)

                if (RealSonar(6) > SecurityThreshold) && (RealSonar(1) <= SecurityThreshold)
                    th = ceil(-pi/2*rand);
                elseif (RealSonar(1) > SecurityThreshold) && (RealSonar(6) <= SecurityThreshold)
                    th = ceil(pi/2*rand);
                else
                    th = ceil (pi*rand);
                end
                %th = 2*pi*rand - xOdomLast(3);
                setdeltaheading(r,rad2deg(th));
                while isheadingdone(r) == 0
                end
            else
                move(r,translation);
                while ismovedone(r) == 0
                end
            end
        case 1         
            th = str2double(inputdlg('Inserire Angolo di rotazione (gradi)', mat2str(xOdomLast)));
            setdeltaheading(r,th);
                while isheadingdone(r) == 1
                end
            ro = str2double(inputdlg('Inserire Spazio', mat2str(xOdomLast)));
            move(r,ro);
                while ismovedone(r) == 1
                end
            pause;
    end
    
   %lock(r);
   
   for i = 2:nParticles
       xP(i,:) = rotate(xP(i,:),th);
       xP(i,:) = traslate(xP(i,:),translation,2);
   end
   
%     xOdomNow = getpose(r);
%     xOdomNow(3) = angle_wrap(deg2rad(xOdomNow(3)));
%     %U contiene il modello del movimento eseguito tra la precedente lettura
%     %odometrica e la attuale lettura
%     u = xOdomNow - xOdomLast;
%     xOdomLast = xOdomNow;
%     
%     xOdom =[xOdom;u];
%     
%     %PREVISIONE, P(x'|u,x)
%      
%     rk = sqrt (u(1)^2 + u(2)^2);
%     u = repmat(u,nParticles,1);
%     
%     u(:,1) = rk * cos(u(:,3) + xP(:,3));
%     u(:,2) = rk * sin(u(:,3) + xP(:,3));
%     
%     u(:,4) = 0;
%     
%     xP = xP + u;
%     xP(:,3) = mod(xP(:,3),2*pi);

    %Lettura sensoriale attraverso ARIA
    RealSonar = readsonar(r);
    %RealSonar ha una struttura [1x8]
    RealSonar=RealSonar(1,1:8);
    %PREVISIONE OSSERVAZIONI
    %Ci mettiamo su ogni particella e valutiamo cosa dovremmo
    %vedere se fossimo in quella locazione
    SimSonar = SonarModel(Map,corners,xP);
    
    Noise = randn(nParticles,8)*eye(8)*100;
    SimSonar = SimSonar.r + Noise;
    %SimSonar = SimSonar.r;
       
    %Devo valutare quanto  differente la lettura sensoriale nella
    %particella da quanto ho realmente in termini percentuali
   
    Innov = abs(repmat(RealSonar,nParticles,1) - SimSonar)./ repmat(RealSonar,nParticles,1);
    %La differenza viene calcolata in termini di valore medio su tutte le
    %letture disponibili
    LeftDiff = mean(Innov(:,1:4),2);
    RightDiff = mean(Innov(:,5:8),2);
    Difference = mean([LeftDiff,RightDiff],2);
    
    %Difference = mean(Innov,2);
    
    %A partire dalla differenza riscontrata associo ad ogni particella un
    %peso proporzionale alla differenza. Maggiore sar la differenza
    %riscontrata minore sar il corrispondente peso associato alla
    %particella
    xP(:,4) = exp(-1*Difference) + 0.001;
    
    switch TypeOfResample 
        case 0
            SelectWithReplacement;
            CheckParticles(xMin,xMax,yMin,yMax);
        case 1
            SelectWithReplacement;
            lambda = 0.5;
            xVar = std(xP);
            Ip = sqrt(xVar(1)^2 + xVar(2)) / sqrt(xMap^2 + yMap^2);
            nParticles =  ceil(nParticles * exp(-lambda * Ip));  %funzione della localizzazione
            ResizeParticlePopulation(nParticles);
            CheckParticles(xMin,xMax,yMin,yMax);
            nPart=[nPart;nParticles];  
    end
    xEst = mean(xP);	
    %xEst = EstimatePosition;
    
    xEstTot = [xEstTot;xEst];
    
    figure(1);
    set(hPoints,'XData',xP(:,1));
    set(hPoints,'YData',xP(:,2));
    
    set(meanXp,'XData', xEst(1));
    set(meanXp,'YData', xEst(2));
    
    %TODO riabilitare secondo plot di seguito che ora ho disabilitato per
    %prestazioni migliori
   % figure(2);plot(xP(1,:),xP(2,:),'.');title('Particle Distribution');
    drawnow;
    if noAvi == 0
        F = getframe(gca);
        mov = addframe(mov,F);
    end
    %unlock(r);
end


if noAvi == 0
    mov = close(mov);
end

%%---%%
%Tecnica di resampling che seleziona le particelle peso maggiore della 
%soglia imposta
function selectParticles (thresold)
global xP;
index = find(xP(:,4)>=thresold);
xP = xP(index,:);

%%---%%
%Metodo di resampling in cui ogni particella ha una probabilit 
%di replica proporzionale al proprio peso
function SelectWithReplacement ()
global xP; global nParticles;
%riselezione basata sui pesi:
%Particelle con un peso pi elevato occuperanno una maggiore percentuale
%delle ordinate di modo che pi valori verranno mappati su quella
%particella
CDF = cumsum(xP(:,4))/sum(xP(:,4));
iSelect  = rand(nParticles,1);
%table lookup
iNextGeneration = interp1(CDF,1:nParticles,iSelect,'nearest','extrap');
%Copio le particelle selezionate nella nuova generazione
xP = xP(iNextGeneration,:);

%%---%%
%Ridimensiona la popolazione di particelle operando un campionamento
%uniforme sulla distribuzione di partenza
function ResizeParticlePopulation(nNewParticles)
global xP;
xPNew = zeros (nNewParticles,4);
nActualParticles = size(xP,1);
for j=1:nNewParticles
    iSelect = ceil(rand * nActualParticles);
    xPNew(j,:) = xP(iSelect,:);
end
xP = xPNew;

%%---%%
%Distribuisce un numero predefinito di particelle attorno ai valori con
%maggiore peso di tutta la popolazione
function DistribuiteParticles (nPartNew)
global xP; global nParticles;

app = zeros(2,3);
[app(1,1),app(2,1)]=max(xP(:,4));
xP(app(2,1),4) = NaN;
[app(1,2),app(2,2)]=max(xP(:,4));
xP(app(2,2),4) = NaN;
[app(1,3),app(2,3)]=max(xP(:,4));
xP(app(2,3),4) = NaN;
sigma = 1000;
sigmaTh = 0.4363; %25

for i=1:nPartNew
    j = ceil(nParticles * rand);
    R = rand;
    if R < 0.7
        z=1;
    elseif R>0.6 && R<0.8
        z=2;
    else
        z=3;
    end
    xP(j,:) = xP(app(2,z),:) + rand(1,4)*diag([sigma,sigma,sigmaTh,0]);
end

%%---%%
%Sostituisce secondo una distribuzione uniforme valori della popolazione
%con valori estratti a partire da tutta la mappa
%n.b. dopo questa funzione il vettore pesi perde senso
function DistribuiteParticleOverMap()
global xP; global nParticles; global Areas;
%def : Areas = [xMin xMax yMin yMax]

nAreas = size(Areas,1);

xMin = Areas(:,1);
xMax = Areas(:,2);
yMin = Areas(:,3);
yMax = Areas(:,4);

AreaTot = sum((xMax-xMin).*(yMax-yMin));

nAreaParticles = ceil(nParticles .* (xMax-xMin).*(yMax-yMin)./AreaTot);

nParticles = sum(nAreaParticles);
%xP matrice delle particelle: [nParticels x 4]
%xP : [nParticles x (x,y,th,peso)]
xP = [];

for i = 1:nAreas     
    newxP = [xMin(i) + (xMax(i)-xMin(i))*rand(nAreaParticles(i),1), ...
             yMin(i) + (yMax(i)-yMin(i))*rand(nAreaParticles(i),1), ...
             2*pi*rand(nAreaParticles(i),1)];
         
    xP = [xP; newxP];
end

%%---%%
%Controlla se ogni particella appartiene ad una delle aree consentite della
%mappa, nel caso negativit la particella viene rigenerata nell'area
%ammissibile
function CheckParticles(xMin,xMax,yMin,yMax)
global xP; global nParticles
nAreas = size(xMin,1);
Threshold = 100;
for i=1:nParticles
    isParticleInArea = 0;
    for j=1:nAreas
        if (xP(i,1) > xMin(j)-Threshold) && (xP(i,1)<xMax(j)+Threshold) && (xP(i,2) > yMin(j)-Threshold) && (xP(i,2)< yMax(j)+Threshold)
            isParticleInArea = 1;
            break;
        end
    end
    if isParticleInArea == 0 %la particella  uscita dall'area ammissibile
        %Ricreo la particella in una delle aree ammissibili scelta a caso
        iSelectArea = ceil(nAreas * rand + 0.00000001);
        xP(i,:) = [xMin(iSelectArea) + (xMax(iSelectArea) - xMin(iSelectArea))*rand, ...
            yMin(iSelectArea) + (yMax(iSelectArea) - yMin(iSelectArea))*rand, ...
            2*pi*rand, 0];
    end
end

function alpha = angle_wrap(alpha)
if(alpha<0 && alpha >=-pi )
    alpha=2*pi-abs(alpha);
end

function xEst = EstimatePosition()
global xP;
total = sum(xP(:,4));
xP(:,4) = xP(:,4)./total;
xEst = sum(xP(:,1:3).* repmat(xP(:,4),1,3));

%%---%%
%Distribuisce le particelle nella mappa facendo si che a partire da ogni
%particella generata si crei una sotto popolazione di nSubParticles
%individui in cui si varia solo l'orientamento
%INPUT
%   nSubParticles, numero di individui per ogni sotto popolazione
%OUTPUT
%   void
%Nota:
%   Radius, indica il raggio entro cui devo generare le particelle a
%   partire da quella originale
function DistribuiteParticlesOrientation(SubAreas, nSubParticles)
global xP; global nParticles;% global Areas;
%def : Areas = [xMin xMax yMin yMax]
Radius = 100;

nAreas = size(SubAreas,1);

xMin = SubAreas(:,1);
xMax = SubAreas(:,2);
yMin = SubAreas(:,3);
yMax = SubAreas(:,4);

AreaTot = sum((xMax-xMin).*(yMax-yMin));

nAreaParticles = ceil(nParticles .* (xMax-xMin).*(yMax-yMin)./AreaTot);

nParticles = sum(nAreaParticles)*nSubParticles;
%xP matrice delle particelle: [nParticels x 4]
%xP : [nParticles x (x,y,th,peso)]
xP = [];
Angle = (2*pi/nSubParticles);

for i = 1:nAreas     
    newxP = [xMin(i) + (xMax(i)-xMin(i))*rand(nAreaParticles(i),1), ...
             yMin(i) + (yMax(i)-yMin(i))*rand(nAreaParticles(i),1), ...
             zeros(nAreaParticles(i),1)];
         AngleVector=[];
    for j = 1:nSubParticles
         AngleVector = [AngleVector; Angle*(j-1)+Angle*rand(nAreaParticles(i),1)];        
    end
    SubParticles = [repmat(newxP(:,1:2),nSubParticles,1) + rand(nSubParticles*nAreaParticles(i),2)*Radius , AngleVector];
    xP = [xP; SubParticles];
end
%%---%%
%Simula la rotazione del robot e il conseguente errore odometrico
%INPUT:
%- pos : variabile contenente la posizione di partenza
%       pos : (1x[x,y,th])
%- th : quantit di radianti di cui devo ruotare
%       th : (1x1) radianti
%OUTPUT:
%- pos : variabile contenente la posizione corrente
%       pos : (1x[x,y,th])
function [pos] = rotate(pos, th)
Mrot = 0.00175; %in radianti
Srot = 0.0175 * 10^-3; %in radianti
pos(3) = pos(3) + th + Mrot + Srot * th *randn;
pos(3) = angle_wrap(pos(3));

%%---%%
%Simula la traslazione del robot con il conseguente errore odometrico
%INPUT:
%- pos : variabile contenente la posizione di partenza
%       pos : (1x[x,y,th])
%- r : distanza da percorrere
%       th : (1x1) mm
%- nSubSteps : numero di sottostep entro cui discretizzare l'operazione di
%       traslazione
%OUTPUT:
%- pos : variabile contenente la posizione corrente
%       pos : (1x[x,y,th])
function [pos] = traslate(pos,r,nSubSteps)
rstep = r/nSubSteps;
%FIXME: da impostare i valori veri per l'odometria
Mtrs = 1 * 10^-3;
Strs = 0.00175;   %dev standard
Mdrft = 1 * 10^-3; 
Sdrft = 0.00175;

for j = 1:nSubSteps
    Etrs = Mtrs * rstep + Strs * rstep * randn;
    Edrft = Mdrft * rstep + Sdrft * rstep * randn;
    pos(3) = pos(3) + Edrft; % orientamento
    pos(3) = angle_wrap(pos(3));
    pos(1) = pos(1) + ((rstep + Etrs) * cos (pos(3))); %x
    pos(2) = pos(2) + ((rstep + Etrs) * sin (pos(3))); %y
    Edrft = Mdrft *rstep + Sdrft *rstep *randn;
    pos(3) = pos(3) + Edrft; % orientamento
    pos(3) = angle_wrap(pos(3));
end