%Monte-Carlo based localization
%----
%Simulatore del movimento di un robot e dell'algoritmo Monte Carlo per 
%la localizzazione.
%IL presente simulatore ha lo scopo di validare l'algoritmo MCL e gli eventuali
%algoritmi di resampling. Gli algoritmi validati vengono poi utilizzati
%nell'applicazione alla localizzazione e la guidadi robot compatibili con
%la dll di interfaccia ARIA-Matlab
%INPUT:
%- map : matrice di linee che descrive la mappa, 
%           map : (nLines x [x1,x2,y1,y2])
%- Corners : Matrice contenente le coordinate dei punti di angolo della
%           mappa, Corners : (nCorners x [x,y])
%- Areas : Matrice contenente la descrizione delle aree entro cui il robot
%           pu localizzarsi, Areas : (nAreas x [x1,y1,x2,y2])
%- NParticles : Numero di particelle massime da impiegare nella
%           localizzazione del robot, NParticles : (1x1)
%- NSteps : Numero di passi massimi per la simulazione, nel caso il robot
%           si localizzi prima di NSteps verr cmq inseguita la posizione 
%           del robot, NSteps : (1x1)
%- TypeOfResample : Indica la tipologia di campionamento da impiegare:
%           0 : Campionamento con sovrascrittura
%           1 : Campionamento con sovrascrittura e ridimensionamento numero
%           di particelle
%           Si rimanda alla documentazione cartacea per i dettagli di ogni
%           tipo di campionamento.
%-xTrue : Posizione di partenza,
%           xTrue : (1x[x,y,th])
%OUTPUT: 
%   void
%----
%Powered by Mancini Adriano & Caponetti Fabio A.A. 2004/2005
%CdL Specialistica in Automazione Industriale
function [xTrue,xEst] = MCLsim_v2(map, Corners, areas, NParticles, nSteps, TypeOfResample,xtrue,AviFileName)

global Map; global corners; global nParticles; global xTrue;
global xP; global Areas;

close all;warning off all;

%Inizializzazione delle variabili globali di progetto. Sono impiegate
%variabili globali per la memorizzazione delle strutture dati comuni a
%tutte le funzioni di modo da ottenere un sensibile miglioramento in
%termini di prestazioni
corners = Corners;
nParticles = NParticles;
Map = map;
xTrue = xtrue;
Areas = areas;

%xTrue posione vera del robot nell'enviroment di simulazione
%xTrue = [500 500 0];

%Inizializzo la lettura odometrica
xOdomLast = getOdometry;

xMap = max(max(Map(:,1:2)));
yMap = max(max(Map(:,3:4)));

%La distribuzione iniziale delle particelle  creata secondo una
%distribuzione di probabilit uniforme
%N.B. L'orientamento del robot  compreso tra [0;2pi]

% nAreas = size(Areas,1);

xMax = max(Areas(:,1),Areas(:,3));
yMax = max(Areas(:,2),Areas(:,4));

xMin = min(Areas(:,1),Areas(:,3));
yMin = min(Areas(:,2),Areas(:,4));

Areas = [ xMin xMax yMin yMax ];
DistribuiteParticlesOrientation(3);

%DistribuiteParticlesOverMap;
%ogni particella ha lo stesso peso d'apprincipio
xP(:,4) = 1/nParticles;

%Inizializzo la grafica
gcf = figure(1); hold on; grid off; axis equal;
DrawMap(Map);hold on;
set(gcf,'doublebuffer','on');
hPoints = plot(xP(:,1),xP(:,2),'.');
meanXp = plot(xTrue(1),xTrue(2),'r*');
TruePos = plot(xTrue(1),xTrue(2),'gX');
drawnow;

noAvi = 0;
if nargin < 8
    noAvi = 1;
else
    try
        mov = avifile(AviFileName);
    catch
        disp ('Problemi nella creazione del file avi...');
        noAvi = 1;
    end
end

for k = 2:nSteps
    %Simulo una iterazione del mondo
    [r,th] = SimulateWorld;
    
    %PREVISIONE
    %Per ogni particella applico il modello del moto al fine di ottenere P(x'|u,x)
    for i = 2:nParticles
        xP(i,:) = rotate(xP(i,:),th);
        xP(i,:) = traslate(xP(i,:),r,2);
    end
    
%     xOdomNow = getOdometry;
%     %U contiene il modello del movimento eseguito tra la precedente lettura
%     %odometrica e la attuale lettura
%     u = xOdomNow - xOdomLast;
%     xOdomLast = xOdomNow;
% 
%     %PREVISIONE, P(x'|u,x)
%     %rk - distanza percorsa secondo la formula di eulero
%     rk = sqrt (u(1)^2 + u(2)^2);
%     %vettorializzo le operazioni di aggiornamento delle particelle
%     u = repmat(u,nParticles,1);
%     TVar = 200;
%     u(:,1) = rk * cos(u(:,3) + xP(:,3)) + randn(nParticles,1) * TVar;
%     u(:,2) = rk * sin(u(:,3) + xP(:,3)) + randn(nParticles,1) * TVar;
% 
%     u(:,4) = 0;
%     %applico il modello del moto appena eseguito dal robot ad ogni
%     %particella
%     xP = xP + u;
%     DVar = 0.0873;
%     xP(:,3) = mod(xP(:,3),2*pi) + randn(nParticles,1) * DVar;

    RealSonar = SonarModel(Map ,corners ,xTrue);
    Noise = randn(1,8) * eye(8)*100;
    RealSonar = RealSonar.r + Noise;
    %RealSonar = ToggleNaN(RealSonar.r,3000);
         
    %predict observation
    %Ci mettiamo su ogni particella e valutiamo cosa dovremmo
    %vedere se fossimo in quella locazione
    SimSonar = SonarModel(Map, corners, xP);
    Noise = randn(nParticles,8) * eye(8)*100;
    SimSonar = SimSonar.r + Noise;
    
    %Devo valutare quanto  differente la lettura sensoriale nella
    %particella da quanto ho realmente in termini percentuali
    
    Innov = abs(repmat(RealSonar,nParticles,1) - SimSonar)./ repmat(RealSonar,nParticles,1);
    %Da un vettore di differenze mi riporto ad uno scalare tenendo conto
    %del lato del robot
    LeftDiff = mean(Innov(:,1:4),2);
    RightDiff = mean(Innov(:,5:8),2);
    Difference = mean([LeftDiff,RightDiff],2);
    %Difference = mean(Innov,2);
    %Associo ad ogni particella un peso funzione della differenza. Maggiore
    % la differenza minore sar il peso, minore sar il peso pi vicino ad
    %uno sar il valore.
    xP(:,4) = exp(-1*Difference) + 0.01;
    
    switch TypeOfResample
        case 0
            SelectWithReplacement;
            CheckParticles;
        case 1
            SelectWithReplacement;
            lambda = 0.5;
            xVar = std(xP);
            Ip = sqrt(xVar(1)^2 + xVar(2)) / sqrt(xMap^2 + yMap^2);
            nParticles =  ceil(nParticles * exp(-lambda * Ip));  %funzione della localizzazione
            ResizeParticlePopulation(nParticles);
            CheckParticles;
            %distribuisco 1/100 di tutta la popolazione attorno ai valori +
            %probabili
            %nPartNew = ceil(nParticles * 1/100);
            %DistribuiteParticles (nPartNew);
            %DistribuiteParticleOverMap(nPartNew,xMap,yMap);
    end

    %xEst = mean(xP);
    
    xEst = EstimatePosition;
    
    %figure(1);
    set(hPoints,'XData',xP(:,1));
    set(hPoints,'YData',xP(:,2));

    set(meanXp,'XData', xEst(1));
    set(meanXp,'YData', xEst(2));

    set(TruePos,'XData', xTrue(1));
    set(TruePos,'YData', xTrue(2));
   
    %TODO: riabilitare secondo plot di seguito che ora ho disabilitato per
    %prestazioni migliori
    % figure(2);plot(xP(1,:),xP(2,:),'.');title('Particle Distribution');
    drawnow;
    if noAvi == 0
        F = getframe(gca);
        mov = addframe(mov,F);
    end
    
end
%La stima della posizione  semplicemente la media di tutte le particelle
x = xEst(1);
y = xEst(2);
th = xEst(3);
if noAvi == 0
    mov = close(mov);
end

%%---%%
%Tecnica di resampling che seleziona le particelle peso maggiore della 
%soglia imposta
%INPUT:
%- thresold : soglia secondo cui selezionare
%           thresold, (1x1) double
%OUTPUT:
% void
function selectParticles (thresold)
global xP;
index = find(xP(:,4) >= thresold);
xP = xP(:,index);

%%---%%
%Simula il movimento del robot all'interno dell'enviroment di simulazione
%INPUT:
% void
%OUTPUT:
%- r : distanza percorsa
%       r, (1x1) double
%- th : variazione dell'orientamento compiuta
%       th, (1x1) radianti
function [r,th] = SimulateWorld()
global xTrue; global Map; global corners;

RealSonar = SonarModel(Map,corners,xTrue);
Noise = randn(1,8) * eye(8)*100;
RealSonar = RealSonar.r + Noise;

type = 0;

%determina la sorgente della azioni da impartire al robot
switch type 
    case 0 %obstacle avoidance automatico
        th = 0;
        r = 0;          
        SecurityThreshold = 500;
        
        if (RealSonar(1) <= SecurityThreshold) || (RealSonar(2) <= SecurityThreshold)||...
                (RealSonar(3) <= SecurityThreshold) || (RealSonar(4) <= SecurityThreshold)||...
                (RealSonar(5) <= SecurityThreshold) || (RealSonar(6) <= SecurityThreshold)||...
                (RealSonar(7) <= SecurityThreshold) || (RealSonar(8) <= SecurityThreshold)
            
            if (RealSonar(1) <= SecurityThreshold)
                th = ceil(-pi/2*rand);
            elseif (RealSonar(6) <= SecurityThreshold)
               th = ceil(pi/2*rand);
            else
                th = ceil (pi*rand);
            end
            xTrue = rotate(xTrue, th);
        else
            r = 200;
            xTrue = traslate(xTrue, r, 2);
        end
                      
    case 1 %input manuale delle azioni
        th = str2double(inputdlg('Inserire Angolo di rotazione (rad)', mat2str(xTrue)));
        r = str2double(inputdlg('Inserire Spazio', mat2str(xTrue)));
        xTrue = rotate(xTrue,th);
        xTrue = traslate(xTrue,r,2);
end

%%---%%
%Simula la rotazione del robot e il conseguente errore odometrico
%INPUT:
%- pos : variabile contenente la posizione di partenza
%       pos : (1x[x,y,th])
%- th : quantit di radianti di cui devo ruotare
%       th : (1x1) radianti
%OUTPUT:
%- pos : variabile contenente la posizione corrente
%       pos : (1x[x,y,th])
function [pos] = rotate(pos, th)
Mrot = 0.00175; %in radianti
Srot = 0.0175 * 10^-3; %in radianti
pos(3) = pos(3) + th + Mrot + Srot * th *randn;
pos(3) = angle_wrap(pos(3));

%%---%%
%Simula la traslazione del robot con il conseguente errore odometrico
%INPUT:
%- pos : variabile contenente la posizione di partenza
%       pos : (1x[x,y,th])
%- r : distanza da percorrere
%       th : (1x1) mm
%- nSubSteps : numero di sottostep entro cui discretizzare l'operazione di
%       traslazione
%OUTPUT:
%- pos : variabile contenente la posizione corrente
%       pos : (1x[x,y,th])
function [pos] = traslate(pos,r,nSubSteps)
rstep = r/nSubSteps;
%FIXME: da impostare i valori veri per l'odometria
Mtrs = 1 * 10^-3;
Strs = 0.00175;   %dev standard
Mdrft = 1 * 10^-3; 
Sdrft = 0.00175;

for j = 1:nSubSteps
    Etrs = Mtrs * rstep + Strs * rstep * randn;
    Edrft = Mdrft * rstep + Sdrft * rstep * randn;
    pos(3) = pos(3) + Edrft; % orientamento
    pos(3) = angle_wrap(pos(3));
    pos(1) = pos(1) + ((rstep + Etrs) * cos (pos(3))); %x
    pos(2) = pos(2) + ((rstep + Etrs) * sin (pos(3))); %y
    Edrft = Mdrft *rstep + Sdrft *rstep *randn;
    pos(3) = pos(3) + Edrft; % orientamento
    pos(3) = angle_wrap(pos(3));
end

%%---%%
%Simula la lettura degli encoder
%INPUT
% void
%OUTPUT
%- xnow : posizione corrente sporcata da rumore 
%         xnow : (1x[x,y,th]) double
function [xnow]= getOdometry()
persistent LastOdom; %internal to robot low-level controller
global xTrue;
if(isempty(LastOdom))
    LastOdom = xTrue;
end;
UTrue = diag([0.0001,0.0001,0.0003046]);
uNoise = randn(1,3) * sqrt(UTrue);
xnow = xTrue + uNoise;
xnow(3) = angle_wrap(xnow(3));
LastOdom = xnow;

%%---%%
%Riconduce da [-pi;pi] a [0;2pi] e riporta tutti gli angoli ad un angolo giro
%INPUT:
%- alpha : angolo da processare
%       alpha : (1x1) radianti
%OUTPUT
%- alpha : angolo processato
%       alpha : (1x1) radianti
function alpha = angle_wrap(alpha)
if(alpha<0 && alpha >=-pi )
    alpha=2*pi-abs(alpha);
end
alpha = mod(alpha,2*pi);

%%---%%
%Metodo di resampling in cui ogni particella ha una probabilit 
%di replica proporzionale al proprio peso
%INPUT
% void
%OUTPUT
% void
function SelectWithReplacement ()
global xP; global nParticles;
%riselezione basata sui pesi:
%Particelle con un peso pi elevato occuperanno una maggiore percentuale
%delle ordinate di modo che pi valori verranno mappati su quella
%particella
CDF = cumsum(xP(:,4))/sum(xP(:,4));
iSelect  = rand(nParticles,1);
%table lookup
iNextGeneration = interp1(CDF,1:nParticles,iSelect,'nearest','extrap');
%Copio le particelle selezionate nella nuova generazione
xP = xP(iNextGeneration,:);

%%---%%
%Ridimensiona la popolazione di particelle operando un campionamento
%uniforme sulla distribuzione di partenza
%INPUT
%- nNewParticles : numero di particelle desiderate
%               nNewParticles : (1x1) integer
%OUTPUT
% void
function ResizeParticlePopulation(nNewParticles)
global xP;
xPNew = zeros (nNewParticles,4);
nActualParticles = size(xP,1);
for j=1:nNewParticles
    iSelect = ceil(rand * nActualParticles);
    xPNew(j,:) = xP(iSelect,:);
end
xP = xPNew;

%%---%%
%Controlla se ogni particella appartiene ad una delle aree consentite della
%mappa, nel caso negativit la particella viene rigenerata nell'area
%ammissibile
%INPUT
%Descrivono i punti delle aree di interesse
%- xMin (nAreas x 1)
%- xMax (nAreas x 1)
%- YMin (nAreas x 1)
%- yMax (nAreas x 1)
%OUTPUT
% void
function CheckParticles()
global xP; global nParticles; global Areas;

xMin = Areas(:,1);
xMax = Areas(:,2);
yMin = Areas(:,3);
yMax = Areas(:,4);

nAreas = size(Areas,1);
Threshold = 100;
for i=1:nParticles
    isParticleInArea = 0;
    for j=1:nAreas
        if (xP(i,1) > xMin(j)-Threshold) && (xP(i,1)<xMax(j)+Threshold) && (xP(i,2) > yMin(j)-Threshold) && (xP(i,2)< yMax(j)+Threshold)
            isParticleInArea = 1;
            break;
        end
    end
    if isParticleInArea == 0 %la particella  uscita dall'area ammissibile
        %Ricreo la particella in una delle aree ammissibili scelta a caso
        iSelectArea = ceil(nAreas * rand + 0.00000001);
        xP(i,:) = [xMin(iSelectArea) + (xMax(iSelectArea) - xMin(iSelectArea))*rand, ...
            yMin(iSelectArea) + (yMax(iSelectArea) - yMin(iSelectArea))*rand, ...
            2*pi*rand, 0];
    end
end

%%---%%
%Distribuisce un numero predefinito di particelle attorno ai valori con
%maggiore peso di tutta la popolazione
%INPUT
%- nPartNew : numero di particelle da distribuire
%           nPartNew (1x1) integer
%OUTPUT
% void
function DistribuiteParticles (nPartNew)
global xP; global nParticles;

app = zeros(2,3);
[app(1,1),app(2,1)]=max(xP(:,4));
xP(app(2,1),4) = NaN;
[app(1,2),app(2,2)]=max(xP(:,4));
xP(app(2,2),4) = NaN;
[app(1,3),app(2,3)]=max(xP(:,4));
xP(app(2,3),4) = NaN;
sigma = 1000;
sigmaTh = 0.4363; %25

for i=1:nPartNew
    j = ceil(nParticles * rand);
    R = rand;
    if R < 0.7
        z=1;
    elseif R>0.6 && R<0.8
        z=2;
    else
        z=3;
    end
    xP(j,:) = xP(app(2,z),:) + rand(1,4)*diag([sigma,sigma,sigmaTh,0]);
end

%%---%%
%Sostituisce secondo una distribuzione uniforme valori della popolazione
%con valori estratti a partire da tutta la mappa
%n.b. dopo questa funzione il vettore pesi perde senso
%INPUT
%- nNewParticles : numero di particelle da distribuire
%           nNewParticles (1x1) integer
%- xMap : limite superiore della mappa lungo le ascisse
%- yMap : limite superiore della mappa lungo le ordinate
%OUTPUT
% void
function DistribuiteParticlesOverMap()
global xP; global nParticles; global Areas;
%def : Areas = [xMin xMax yMin yMax]

nAreas = size(Areas,1);

xMin = Areas(:,1);
xMax = Areas(:,2);
yMin = Areas(:,3);
yMax = Areas(:,4);

AreaTot = sum((xMax-xMin).*(yMax-yMin));

nAreaParticles = ceil(nParticles .* (xMax-xMin).*(yMax-yMin)./AreaTot);

nParticles = sum(nAreaParticles);
%xP matrice delle particelle: [nParticels x 4]
%xP : [nParticles x (x,y,th,peso)]
xP = [];

for i = 1:nAreas     
    newxP = [xMin(i) + (xMax(i)-xMin(i))*rand(nAreaParticles(i),1), ...
             yMin(i) + (yMax(i)-yMin(i))*rand(nAreaParticles(i),1), ...
             2*pi*rand(nAreaParticles(i),1)];
         
    xP = [xP; newxP];
end

%%---%%
%Stima la posizione attraverso una media pesata delle particelle
%INPUT
% void
%OUTPUT
%- xEst : posizione stimata nella mappa
%       xEst : (1x[x,y,th])
function xEst = EstimatePosition()
global xP;
total = sum(xP(:,4));
xP(:,4) = xP(:,4)./total;
xEst = sum(xP(:,1:3).* repmat(xP(:,4),1,3));

%%---%%
%Distribuisce le particelle nella mappa facendo si che a partire da ogni
%particella generata si crei una sotto popolazione di nSubParticles
%individui in cui si varia solo l'orientamento
%INPUT
%   nSubParticles, numero di individui per ogni sotto popolazione
%OUTPUT
%   void
%Nota:
%   Radius, indica il raggio entro cui devo generare le particelle a
%   partire da quella originale
function DistribuiteParticlesOrientation(nSubParticles)
global xP; global nParticles; global Areas;
%def : Areas = [xMin xMax yMin yMax]
Radius = 300;

nAreas = size(Areas,1);

xMin = Areas(:,1);
xMax = Areas(:,2);
yMin = Areas(:,3);
yMax = Areas(:,4);

AreaTot = sum((xMax-xMin).*(yMax-yMin));

nAreaParticles = ceil(nParticles .* (xMax-xMin).*(yMax-yMin)./AreaTot);

nParticles = sum(nAreaParticles)*nSubParticles;
%xP matrice delle particelle: [nParticels x 4]
%xP : [nParticles x (x,y,th,peso)]
xP = [];
Angle = (2*pi/nSubParticles);

for i = 1:nAreas     
    newxP = [xMin(i) + (xMax(i)-xMin(i))*rand(nAreaParticles(i),1), ...
             yMin(i) + (yMax(i)-yMin(i))*rand(nAreaParticles(i),1), ...
             zeros(nAreaParticles(i),1)];
         AngleVector=[];
    for j = 1:nSubParticles
         AngleVector = [AngleVector; Angle*(j-1)+Angle*rand(nAreaParticles(i),1)];        
    end
    SubParticles = [repmat(newxP(:,1:2),nSubParticles,1) + rand(nSubParticles*nAreaParticles(i),2)*Radius , AngleVector];
    xP = [xP; SubParticles];
end

