%Mancini Adriano & Caponetti Fabio
%Montecarlo Localization for Mobile Robots (AmigoBot or others)
%questa funzione implementa il modello del sonar permettendo di ottenere 
%una conoscenza apriori di ci che ci si aspetta di misurare essendo in una
%particolare posizione
%Parametri input: Punti che caratterizzano le linee, corner e serie delle
%particelle (necessari entrambi)
%
%Parametri output: Misure secondo il modello e validit delle stesse

function [z]=SonarModel2(room,room2,X)

%Parametri Sonar globali perch usati in molte funzioni
global beta;
global R;
global sonar;
global walls;
global corner;
sonar=zeros(8,4);
%Pioneer 3 dx
sonar=[69, 136, deg2rad(90);
114, 119, deg2rad(50);
148, 78, deg2rad(30);
166, 27, deg2rad(10);
166, -27, deg2rad(-10);
148, -78, deg2rad(-30);
114, -119, deg2rad(-50);
69, -136, deg2rad(-90)];
%AmigoBot
% sonar = [
%     76 100 deg2rad(90) ;
%     125 75 deg2rad(41) ;
%     150 30 deg2rad(15) ;
%     150 -30 deg2rad(-15) ;
%     125 -75 deg2rad(-41) ;
%     76 -100 deg2rad(-90);
%     -140 -58 deg2rad(-145) ;
%     -140 58 deg2rad(145) ];
beta=deg2rad(30);
R=5000;
walls=room;
corner=room2;
%Necessario inserire posizione ed angolo di ogni sonar rispetto alla terna
%solidale con il robot

%Parametri relativi all'algoritmo
n_particles=length(X(:,1));
n_sonar=8;

z=struct('r',{},'alpha',{},'flag',{});
z=[];
for j=1:1:n_particles
    for i=1:1:n_sonar
        
        [r,alpha,flag]=search_obstacle(X(j,:),i);          
        z.r(j,i)=r;
        z.alpha(j,i)=alpha;
        z.flag(j,i)=flag;
        
    end
end

%la function model prende come parametri in ingresso
%il tipo di ostacolo muro o angolo, i parametri necessari per il calcolo
%degli output che in entrambe i casi sono due, ed il numero del sonar che
%deve essere stimato
%flag=1 se ok else 0
function [r,alpha,flag]=model(par1,par2,particle,i)
global sonar;
global beta;
global R;

r=sqrt((par1-(sonar(i,1)+particle(1,1)))^2+(par2-(sonar(i,2)+particle(1,2)))^2);
alpha=atan2(par2-(sonar(i,2)+particle(1,2)),par1-(sonar(i,1)+particle(1,1)));
%necessario fare il wrap
alpha=angle_wrap(alpha);
if(alpha>(particle(1,3)+sonar(i,3))-beta/2 && alpha<(particle(1,3)+sonar(i,3))+beta/2 && r<R)
    flag=1;
else
    flag=0;
end

function alpha=angle_wrap(alpha)
if(alpha>=0 && alpha <=pi )
    alpha=alpha;
else
    alpha=2*pi-abs(alpha);
end
%Calcola Range Cerca ostacolo ed individua il tipo possono forse esssere
%implementate da questo punto in avanti
%questa funzione ha il compito di inviduare i punti limite che descrivono
%la regione ammissibile
%mode indica l'approssimazione della'area 
%con 2 / 3 punti per il momento solo 2 punti



function [r,alpha,flag]=search_obstacle(particle,sonar_number)
global walls;
global corner;
global R;
index=[1:length(walls)]';
index2=[1:length(corner)]';


%passo successivo  quello di discriminare tra tutte e scegliere quella che
%minimizza ||Pi-X||
n_index=length(index);
n_index2=length(index2);
dati_tmp=[];
for i=1:1:n_index
    %calcolo punti candidati che minimizzano la distanza dalla
    %particella; successivamente controllo se il punto appartiene alla
    %regione desiderata come [x,y]
    punto(i,:)=init_measure(index(i,1),particle,sonar_number); %da controllare index(1,i) o (i,1)
    %se il punto  valido allora calcolo distanza e angolo mendiante
    %apposita funzione
    if(isnan(punto(i,1))==0)
        [dati_tmp(i,1),dati_tmp(i,2),dati_tmp(i,3)]=model(punto(i,1),punto(i,2),particle,sonar_number);
    else
        [dati_tmp(i,1:3)]=[R,NaN,0];
        %prendo il minimo dei valori validi ed il corrispondente indice
        
    end
end
for i=1+n_index:n_index2+n_index
    [dati_tmp(i,1),dati_tmp(i,2),dati_tmp(i,3)]=model(corner(i-n_index,1),corner(i-n_index,2),particle,sonar_number);
    
end


if(length(dati_tmp)<1)
    r=R;
    alpha=NaN;
    flag=0;
else
    cond=find(dati_tmp(:,3)>0);
    if (length(cond)<1)
        r=R;
        alpha=NaN;
        flag=0;
    else
        %prendo angolo mediante indice index2 e seconda colonna
        [r,index2]=min(dati_tmp(cond,1));
        alpha=dati_tmp(cond(index2),2);
        flag=dati_tmp(cond(index2),3);
    end  
    
end


%Formato dei dati x1x2y1y2

function punto=init_measure(index,particle,i)
global sonar;
global walls;
%riporto il delta x e il delta y rispetto al sistema solidale con il sonar
delta_x=(walls(index,1)-particle(1,1)-sonar(i,1))-(walls(index,2)-particle(1,1)-sonar(i,1));
delta_y=(walls(index,3)-particle(1,2)-sonar(i,2))-(walls(index,4)-particle(1,2)-sonar(i,2));

lambda=-(delta_x*(walls(index,2)-particle(1,1)-sonar(i,1))+delta_y*(walls(index,4)-particle(1,2)-sonar(i,2)))/(delta_x^2+delta_y^2);
if (lambda<0 || lambda >1)
    punto(1,1:2)=[NaN,NaN];
else
    %calcolo del punto sulla retta lambda*P1+ (1-lambda)*P2
    %ovvero punto a minima distanza dalla particle (sistema riferimento
    %centrato sulla particle dunque necessario portarlo nella terna solidale
    %con il particolare sonar
    punto=lambda.*([walls(index,1),walls(index,3)])+(1-lambda).*([walls(index,2),walls(index,4)]);
    %riferisco al sistema globale
    %punto(1,:)=punto_tmp(1,:)%+[particle(1,1)+sonar(i,1),particle(1,2)+sonar(i,2)];
    %misura valida e dunque posso calcolare la distanza e l'angolo
end    